import collections
from datetime import datetime, timezone, timedelta
import enum
from typing import Dict, List, Optional, Tuple

import statistics

from rich.text import Text
from rich.columns import Columns

from pysmelt.rc import SmeltRcHolder
from rich.table import Table
from typing_extensions import cast
from rich.console import Group, RenderableType
from rich.tree import Tree
import rich
from pysmelt.interfaces.target import SmeltTargetType, Target
from pysmelt.output import smelt_console
from dataclasses import dataclass, field
from pysmelt.proto.smelt_telemetry import (
    CommandEvent,
    CommandFinished,
    CommandProfile,
    CommandStdout,
    Event,
)
import betterproto
from rich.progress import (
    Progress,
    RenderableColumn,
    SpinnerColumn,
    Task,
    TaskID,
    TimeElapsedColumn,
)


class Status(enum.Enum):
    scheduled = "scheduled"
    started = "started"
    finished = "finished"
    cancelled = "cancelled"
    skipped = "skipped"


@dataclass
class StatusObj:
    name: str
    started_time: datetime
    command_type: SmeltTargetType
    status: Status


def format_time(total_seconds: float, rich_conformant: bool = False) -> str:
    hours, minutes = divmod(total_seconds, 3600)
    minutes, seconds = divmod(total_seconds, 60)
    time_str = ""
    if rich_conformant:
        return "{:01}:{:02}:{:02}".format(int(hours), int(minutes), int(seconds))

    if hours > 0:
        time_str += f"{int(hours)}h "
    if minutes > 0:
        time_str += f"{int(minutes)}m "
    if seconds > 0:
        time_str += f"{seconds:.3f}s "

    return time_str.strip()


def stringify_mem(num_bytes: float):
    """
    this function will convert bytes to MB.... GB... etc
    """
    for x in ["B", "KB", "MB", "GB", "TB"]:
        if num_bytes < 1024.0:
            return "%3.1f %s" % (num_bytes, x)
        num_bytes /= 1024.0
    return ""


class RenderableTree(RenderableColumn):
    """Tasks that are running"""

    def render(self, task: Task):
        """Show data completed."""
        status_dict: Dict[str, Status] = task.fields["fields"]["status"]
        profile_dict: Dict[str, CommandProfile] = task.fields["fields"]["profile"]
        timedict: Dict[str, datetime] = task.fields["fields"]["start"]

        just_size = 11
        outer_just = 59
        inner_just = 55

        def get_total_tests():
            finished_keys = [k for k, v in status_dict.items() if v == Status.finished]
            num_finished_keys = len(finished_keys)
            total_keys = len(status_dict)
            return Text(
                f"{num_finished_keys}/{total_keys} executed".ljust(outer_just),
                style="bold",
                justify="left",
            )

        total_tests = get_total_tests()

        def get_elapsed():
            elapsed = task.finished_time if task.finished else task.elapsed
            if elapsed is None:
                return Text("-:--:--", style="progress.elapsed")
            delta = timedelta(seconds=max(0, int(elapsed)))
            return Text(str(delta), style="progress.elapsed")

        elapsed = get_elapsed()

        def get_agg(task: Task):
            profile_dict: Dict[str, CommandProfile] = task.fields["fields"]["profile"]
            status_dict: Dict[str, Status] = task.fields["fields"]["status"]
            max_load_and_mem: List[float] = task.fields["fields"]["max"]

            mem = 0.0
            load = 0.0
            for command, status in status_dict.items():
                profile = profile_dict[command] if command in profile_dict else None
                if status == Status.started:
                    if profile:

                        mem += profile.memory_used
                        load += profile.cpu_load
            if mem > max_load_and_mem[0]:
                max_load_and_mem[0] = mem
            if load > max_load_and_mem[1]:
                max_load_and_mem[1] = load
            memstr = stringify_mem(mem)

            return Text(f"{memstr.ljust(just_size)} " + f"{load:.2f}".ljust(just_size))

        root = Tree(
            Columns(("Total".ljust(outer_just), get_agg(task), elapsed), align="left")
        )
        for command, status in status_dict.items():
            profile = profile_dict[command] if command in profile_dict else None
            if status == Status.started:
                execution_time = datetime.now(timezone.utc) - timedict[command]
                total_seconds = execution_time.total_seconds()
                timestr = format_time(total_seconds, rich_conformant=True)
                name = Text(
                    f"Running {command}".ljust(inner_just),
                )

                time = Text(f"{timestr}")

                if profile:
                    mem = stringify_mem(profile.memory_used)
                    cpu_load = f"{profile.cpu_load:.2f}"
                    sub1 = root.add(
                        Columns(
                            [
                                name,
                                Text(f"{mem}".ljust(just_size)),
                                Text(f"{cpu_load}".ljust(just_size)),
                                time,
                            ],
                            align="left",
                        )
                    )

                else:
                    sub1 = root.add(Columns([name, "".ljust(just_size * 2 + 1), time]))

        topper = Columns(
            (
                total_tests,
                "memory".ljust(just_size),
                "cpu load".ljust(just_size),
                "elapsed".ljust(just_size),
            )
        )
        return Columns((topper, root), column_first=True)


def get_logs_from_command_name(command_name: str, num_lines_tail: int = 3) -> str:
    root = SmeltRcHolder.current_rc().smelt_root
    try:
        with open(f"{root}/smelt-out/{command_name}/command.out", "r") as f:
            lines = collections.deque(f, num_lines_tail)
        return "\n".join(lines)
    except Exception as e:
        return f"Could not read {command_name}'s log"


@dataclass
class OutputConsole:
    """
    Simple subscriber for creating a console
    """

    print_stdout: bool = False
    is_done: bool = False
    total_run: int = 0
    total_executing: int = 0
    total_passed: int = 0
    total_tests_passed: int = 0
    total_tests_failed: int = 0
    total_scheduled: int = 0
    status_dict: Dict[str, Status] = field(default_factory=dict)
    profile_dict: Dict[str, CommandProfile] = field(default_factory=dict)
    progress: Optional[Progress] = None
    task: Optional[TaskID] = None
    finished_list: List[Tuple[CommandFinished, str, datetime]] = field(
        default_factory=list
    )
    max_load_and_mem: List[float] = field(default_factory=list)
    skipped_list: List[str] = field(default_factory=list)

    start_time: Dict[str, datetime] = field(default_factory=dict)

    def __enter__(self):

        self.progress = Progress(
            RenderableTree(),
        )
        self.max_load_and_mem = [0, 0]

        self.progress.start()
        self.task = self.progress.add_task(
            "Executing smelt...",
            fields={
                "status": self.status_dict,
                "profile": self.profile_dict,
                "max": self.max_load_and_mem,
                "start": self.start_time,
            },
        )

        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.progress:
            self.progress.stop()
            self.progress = None

        smelt_console.log(
            f"Executed {self.total_run} commands, {self.total_passed} commands passed"
        )

        table = Table(show_header=True, header_style="bold magenta")
        table.add_column("Command Name")
        table.add_column("Status")
        table.add_column("Execution Time")

        new_finished_list = [
            (obj, command_name, end_time - self.start_time[command_name])
            for obj, command_name, end_time in self.finished_list
        ]
        # toggle this if we want to show more or less
        topn = 10
        for obj, command_name, execution_time in sorted(
            new_finished_list, key=lambda x: x[2], reverse=True
        )[:topn]:

            total_seconds = execution_time.total_seconds()
            time_str = format_time(total_seconds)

            table.add_row(
                command_name,
                (
                    "PASSED"
                    if obj.outputs.exit_code == 0
                    else f"FAILED, code: {obj.outputs.exit_code}"
                ),
                time_str,
            )

        if len(new_finished_list) > topn:
            unseen = len(new_finished_list) - topn
            table.add_row(f"and {unseen} other commands...", "")
        smelt_console.log(table)
        fail_table = Table(
            show_header=True, title="Failed commands", header_style="bold magenta"
        )
        fail_table.add_column("Command Name")
        fail_table.add_column("log tail")

        one_failed = False
        for obj, command_name, execution_time in filter(
            lambda x: x[0].outputs.exit_code != 0,
            sorted(new_finished_list, key=lambda x: x[2], reverse=True),
        ):
            one_failed = True
            logs = get_logs_from_command_name(command_name)
            fail_table.add_row(command_name, logs)
        if one_failed:
            smelt_console.log(fail_table)

        failed = self.total_run - self.total_passed
        smelt_console.print(f"[green] {self.total_tests_passed} tests passed ")
        if len(self.skipped_list) != 0:
            smelt_console.print(f"[red] {len(self.skipped_list)} commands skipped")
        if failed != 0:
            smelt_console.print(
                f"[red] {self.total_tests_failed} tests failed, {failed} commands failed"
            )
        execution_times = [
            x[2].total_seconds()
            for x in new_finished_list
            if x[0].command_type == SmeltTargetType.Test.value
        ]

        try:
            average_time = sum(execution_times) / len(execution_times)
            stdev_time = statistics.stdev(execution_times)
            smelt_console.print(
                f"avg test duration {average_time:.2f}s, stddev {stdev_time:.3f}s"
            )
        except Exception as _:
            pass

    def process_message(self, message: Event):
        (variant, event_payload) = betterproto.which_one_of(message, "et")
        if variant == "command":
            event_payload: CommandEvent
            name = event_payload.command_ref
            (command_name, payload) = betterproto.which_one_of(
                event_payload, "CommandVariant"
            )

            if command_name != "stdout" and command_name != "profile":
                self.status_dict[name] = Status(command_name)
                if command_name == "scheduled":
                    self.total_scheduled += 1
                if command_name == "started":
                    self.processed_started(name, message.time)

                if command_name == "finished":
                    payload = cast(CommandFinished, payload)
                    self.process_finished(payload, name, message.time)

                if command_name == "skipped":
                    self.process_skipped(name)
            if command_name == "profile":
                payload = cast(CommandProfile, payload)
                self.profile_dict[name] = payload

            # we are processing stdout of a command
            else:
                if self.progress and self.print_stdout:
                    payload = cast(CommandStdout, payload)
                    self.progress.print(payload.output)

    def processed_started(self, name: str, time: datetime):
        self.total_executing += 1
        self.total_run += 1
        self.start_time[name] = time

    def process_skipped(self, name: str):
        self.skipped_list.append(name)

    def process_finished(self, obj: CommandFinished, command_name: str, time: datetime):
        self.total_executing -= 1
        if obj.outputs.exit_code == 0:
            self.total_passed += 1
            if obj.command_type == SmeltTargetType.Test.value:
                self.total_tests_passed += 1
        elif (
            obj.outputs.exit_code != 0
            and obj.command_type == SmeltTargetType.Test.value
        ):
            self.total_tests_failed += 1
        else:
            pass
        self.finished_list.append((obj, command_name, time))

    def reset(self):
        self.is_done = False



from importlib.util import module_from_spec, spec_from_file_location
from pathlib import Path
import sys
from typing import List
from pysmelt.importer import import_procedural_testlist


from pysmelt.interfaces.target import Target


import contextlib


from pysmelt.path_utils import get_git_root

_import_depth = 0


def get_import_depth() -> int:
    return _import_depth


def inc_import_depth():
    global _import_depth
    _import_depth += 1


def dec_import_depth():
    global _import_depth
    _import_depth -= 1


@contextlib.contextmanager
def capture_targets():
    """
    Patch the __post_init__ method of `Target` to append new instances of
    Target to the yielded list.
    """
    instances: List[Target] = []

    original_init = Target.__post_init__

    def new_init(self: Target, *args, **kwargs):
        original_init(self, *args, **kwargs)
        if _import_depth == 0:
            instances.append(self)

    Target.__post_init__ = new_init

    yield instances

    Target.__post_init__ = original_init

    return


def get_procedural_targets(py_path: str) -> List[Target]:

    with capture_targets() as target_instances:
        # Import procedural testlist py_path for the side effect of adding
        # instances of Target to target_instances.
        import_procedural_testlist(py_path)

    return target_instances


def init_local_rules():
    root_dir = Path(get_git_root()) / "smelt_rules"
    if root_dir.exists():
        for py_file in root_dir.glob("*.py"):
            module_name = py_file.stem  # get filename without extension
            spec = spec_from_file_location(module_name, str(py_file))
            module = module_from_spec(spec)
            spec.loader.exec_module(module)
            sys.modules[module_name] = (
                module  # add the module to the list of globally imported modules
            )



import pathlib
from dataclasses import dataclass
from typing import Dict, Any, Iterable, Optional, Tuple, Type, List

import yaml
from pydantic import BaseModel

from pysmelt.generators.procedural import get_procedural_targets
from pysmelt.importer import (
    DocumentedTarget,
    get_all_targets,
    get_default_targets,
)
from pysmelt.interfaces import Target, Command
from pysmelt.interfaces.paths import (
    SmeltPath,
    TempTarget,
    SmeltPathFetcher,
    local_smelt_path_fetcher,
)
from pysmelt.rc import SmeltRC, SmeltRcHolder
from pysmelt.tracker import ImportTracker


class SerYamlTarget(BaseModel):
    name: str
    rule: str
    rule_args: Dict[str, Any] = {}


@dataclass
class PreTarget:
    target_typ: Type[Target]
    rule_args: Dict[str, Any]


def populate_rule_args(
    target_name: str,
    rule_payload: SerYamlTarget,
    all_rules: Dict[str, DocumentedTarget],
) -> PreTarget:
    rule_payload.rule_args["name"] = target_name
    if rule_payload.rule not in all_rules:

        # TODO: make a pretty error that
        #
        # Says that no rule is visible
        # Prints out all rules that are visible
        # Point to the location where end users can create new rules
        raise RuntimeError(f"Rule named {rule_payload.rule} has not been created!")
    target_type = all_rules[rule_payload.rule]["target"]
    return PreTarget(target_typ=target_type, rule_args=rule_payload.rule_args)


def to_target(pre_target: PreTarget) -> Target:
    return pre_target.target_typ(**pre_target.rule_args)


def get_targets(
    test_list: SmeltPath,
    default_rules_only: bool = False,
    file_fetcher: Optional[SmeltPathFetcher] = None,
) -> Dict[str, Target]:

    abs_path = test_list.to_abs_path()

    if pathlib.Path(abs_path).suffix == ".py":
        if file_fetcher is not None:
            raise RuntimeError(
                "Custom Smelt file fetchers are not yet supported for procedural targets"
            )
        targets = get_procedural_targets(abs_path)

        return {target.name: target for target in targets}

    else:
        if file_fetcher is None:
            yaml_content = local_smelt_path_fetcher(test_list)
        else:
            yaml_content = file_fetcher(test_list)
        return smelt_contents_to_targets(
            yaml_content, default_rules_only=default_rules_only
        )


def parse_smelt(
    test_list: SmeltPath,
    default_rules_only: bool = False,
    file_fetcher: Optional[SmeltPathFetcher] = None,
) -> Tuple[Dict[str, Target], List[Command]]:
    """
    Parses a single test list, and returns the targets and generated commands from the test list


    """
    test_list_orig = test_list
    targets = get_targets(test_list, default_rules_only, file_fetcher=file_fetcher)
    ImportTracker.imported_targets[ImportTracker.local_file_alias()] = targets
    ImportTracker.imported_targets[test_list] = targets

    command_list = lower_targets_to_commands(
        targets.values(), str(pathlib.Path(test_list_orig.to_abs_path()).parent)
    )

    ImportTracker.imported_commands[test_list] = command_list
    return targets, command_list


@dataclass
class SmeltUniverse:
    """
    The smelt "universe" contains all commands that have been generated by parsing a test list
    mapping them to the test lists that they were generated from

    """

    top_file: SmeltPath
    commands: Dict[SmeltPath, List[Command]]

    @property
    def all_commands(self) -> List[Command]:
        return [
            command
            for command_list in self.commands.values()
            for command in command_list
        ]

    @property
    def top_level_commands(self) -> List[Command]:
        return self.commands[self.top_file]


def create_universe(
    starting_file: SmeltPath,
    default_rules_only: bool = False,
    file_fetcher: Optional[SmeltPathFetcher] = None,
) -> SmeltUniverse:
    # Initialize the top file, seen files, visible files, and all commands
    top_file = starting_file
    seen_files = {starting_file}
    visible_files = {starting_file}
    all_commands = {}

    # Parse the "initial" file under consideration and all of the testlists seen to visible files
    targets, commands = parse_smelt(
        starting_file, default_rules_only, file_fetcher=file_fetcher
    )
    for comm in commands:
        for dep in comm.dependencies:
            tt = TempTarget.parse_string_smelt_target(dep, starting_file.path)
            visible_files.add(tt.file_path)
    all_commands[top_file] = commands
    all_commands.update(ImportTracker.imported_commands)
    ImportTracker.clear()

    # Determine new files that are visible but not yet parsed
    new_files = visible_files - seen_files

    # Continue to parse new files until we've seen everything
    while True:
        if len(new_files) != 0:
            file = new_files.pop()
            seen_files.add(file)
            _, new_commands = parse_smelt(
                file,
                default_rules_only,
                file_fetcher=file_fetcher,
            )

            # Add dependencies of new commands to new files if not seen
            for command in new_commands:
                for dep in command.dependencies:
                    tt = TempTarget.parse_string_smelt_target(dep, file.path)
                    if tt.file_path not in seen_files:
                        new_files.add(tt.file_path)

            all_commands[file] = new_commands
            all_commands.update(ImportTracker.imported_commands)
            ImportTracker.clear()

        else:
            # Break the loop if there are no new files
            break

    # Should have everything in the Universe now
    return SmeltUniverse(top_file=top_file, commands=all_commands)


def target_to_command(
    target: Target, working_dir: str
) -> Tuple[Command, Optional[Command], Optional[Command]]:
    rc = SmeltRcHolder.current_rc()
    return (
        target.to_command(working_dir),
        target.to_rebuild_command(working_dir),
        target.to_rerun_command(working_dir),
    )


def lower_targets_to_commands(targets: Iterable[Target], path: str) -> List[Command]:
    return [
        command
        for target in targets
        for command in target_to_command(target, path)
        if command
    ]


def smelt_contents_to_targets(
    smelt_content: str,
    rc: SmeltRC = SmeltRcHolder.current_rc(),
    default_rules_only: bool = False,
) -> Dict[str, Target]:
    rule_inst = yaml.safe_load(smelt_content)
    # NOTE: semantically we split up validation of the smelt file -> converting to target objects -> generating a command list
    # while dependency based
    if default_rules_only:
        all_rules = get_default_targets(rc)
    else:
        all_rules = get_all_targets(rc)
    yaml_targets = [SerYamlTarget(**target) for target in rule_inst]
    pre_targets = {
        target.name: populate_rule_args(target.name, target, all_rules)
        for target in yaml_targets
    }
    return {name: to_target(pre_target) for name, pre_target in pre_targets.items()}



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# sources: client.data.proto
# plugin: python-betterproto
# This file has been @generated

from dataclasses import dataclass
from typing import (
    Dict,
    List,
    Optional,
)

import betterproto


class ProfilingSelection(betterproto.Enum):
    DISABLED = 0
    SIMPLE_PROF = 1
    """only memory and cpu"""


class RunMode(betterproto.Enum):
    Local = 0
    Remote = 1


@dataclass(eq=False, repr=False)
class ClientCommand(betterproto.Message):
    setter: "SetCommands" = betterproto.message_field(1, group="ClientCommands")
    runone: "RunOne" = betterproto.message_field(2, group="ClientCommands")
    runtype: "RunType" = betterproto.message_field(3, group="ClientCommands")
    runmany: "RunMany" = betterproto.message_field(4, group="ClientCommands")
    getcfg: "GetConfig" = betterproto.message_field(5, group="ClientCommands")


@dataclass(eq=False, repr=False)
class SetCommands(betterproto.Message):
    command_content: str = betterproto.string_field(1)
    maybe_def_path: str = betterproto.string_field(2)


@dataclass(eq=False, repr=False)
class RunOne(betterproto.Message):
    command_name: str = betterproto.string_field(1)


@dataclass(eq=False, repr=False)
class RunMany(betterproto.Message):
    command_names: List[str] = betterproto.string_field(1)


@dataclass(eq=False, repr=False)
class RunType(betterproto.Message):
    typeinfo: str = betterproto.string_field(1)
    """
    Eventually, perhaps we should encode this as info in protobuf not today
    babey
    """


@dataclass(eq=False, repr=False)
class GetConfig(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class ClientResp(betterproto.Message):
    """Responses to the client command"""

    current_cfg: "ConfigureSmelt" = betterproto.message_field(
        1, group="ClientResponses"
    )


@dataclass(eq=False, repr=False)
class ConfigureSmelt(betterproto.Message):
    """
    This configuration is done once, when SMELT is initialized The client
    should provide this when creating an smelt handle
    """

    smelt_root: str = betterproto.string_field(1)
    """Should be an absolute path"""

    job_slots: int = betterproto.uint64_field(2)
    """
    number of slots the entire executor has -- analogous to job slots in make
    """

    prof_cfg: "ProfilerCfg" = betterproto.message_field(3)
    """configures how we profile commands"""

    test_only: bool = betterproto.bool_field(4)
    """If true, we ignore the non test commands"""

    silent: bool = betterproto.bool_field(5)
    """If true, we do not transmit stdout from the server"""

    prepare_workspace: bool = betterproto.bool_field(6)
    """
    if true, smelt execution will prepare the workspace to run tests in another
    stage this is useful for splitting the work between a build and test stage
    """

    sandbox_env: bool = betterproto.bool_field(7)
    """
    if true, all commands will have their environment variables sandboxed
    """

    local: "CfgLocal" = betterproto.message_field(10, group="InitExecutor")
    docker: "CfgDocker" = betterproto.message_field(11, group="InitExecutor")
    slurm: "CfgSlurm" = betterproto.message_field(12, group="InitExecutor")


@dataclass(eq=False, repr=False)
class ProfilerCfg(betterproto.Message):
    prof_type: "ProfilingSelection" = betterproto.enum_field(1)
    """if we enable simple profiling"""

    sampling_period: int = betterproto.uint64_field(2)


@dataclass(eq=False, repr=False)
class CfgLocal(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class CfgSlurm(betterproto.Message):
    creds: "AwsCreds" = betterproto.message_field(1)
    maybe_info: "ServerInfo" = betterproto.message_field(2)
    none: bool = betterproto.bool_field(10, group="SealedWorkspace")
    dockerws: "DockerWorkspace" = betterproto.message_field(11, group="SealedWorkspace")


@dataclass(eq=False, repr=False)
class ServerInfo(betterproto.Message):
    """
    Populate info for the slurm server  This is configured from the end-user
    side.  if these data is not provided, then smelt will execute `whoami` to
    derive hostname and will let the OS select the port
    """

    hostname: str = betterproto.string_field(1)
    """this is the hostname used by smelt and all workers, by default"""

    port: int = betterproto.uint32_field(2)
    """
    This is port supplied to each worker a. Sometimes you'll have a proxy in
    front of your smelt server to make deployment easier. e.g.  this is
    practical for  deploying in non-LAN networking environments
    """


@dataclass(eq=False, repr=False)
class AwsCreds(betterproto.Message):
    """
    credentials required to create an S3 Client and upload artifacts to it
    """

    key_id: str = betterproto.string_field(1)
    key: str = betterproto.string_field(2)
    bucket: str = betterproto.string_field(3)
    key_base_path: str = betterproto.string_field(4)


@dataclass(eq=False, repr=False)
class DockerWorkspace(betterproto.Message):
    container_name: str = betterproto.string_field(1)
    workspace_smelt_root: str = betterproto.string_field(2)
    docker_args: List[str] = betterproto.string_field(3)


@dataclass(eq=False, repr=False)
class Ulimit(betterproto.Message):
    """Clone of Bollard proto message Ulimit"""

    name: Optional[str] = betterproto.string_field(1, optional=True, group="_name")
    soft: Optional[int] = betterproto.int64_field(2, optional=True, group="_soft")
    hard: Optional[int] = betterproto.int64_field(3, optional=True, group="_hard")


@dataclass(eq=False, repr=False)
class CfgDocker(betterproto.Message):
    image_name: str = betterproto.string_field(1)
    additional_mounts: Dict[str, str] = betterproto.map_field(
        2, betterproto.TYPE_STRING, betterproto.TYPE_STRING
    )
    ulimits: List["Ulimit"] = betterproto.message_field(3)
    mac_address: Optional[str] = betterproto.string_field(
        4, optional=True, group="_mac_address"
    )
    run_mode: "RunMode" = betterproto.enum_field(5)
    artifact_bind_directory: str = betterproto.string_field(6)



from dataclasses import dataclass

from typing import ClassVar, Dict, List, Optional


from typing import TYPE_CHECKING

from pysmelt.interfaces.paths import SmeltPath, TempTarget


if TYPE_CHECKING:

    from pysmelt.interfaces.command import Command
    from pysmelt.interfaces.target import Target, TargetRef


@dataclass
class ImportTracker:
    imported_commands: ClassVar[Dict[SmeltPath, List["Command"]]] = {}
    imported_targets: ClassVar[Dict[SmeltPath, Dict[str, "Target"]]] = {}

    @staticmethod
    def clear():
        ImportTracker.imported_commands = {}

    @staticmethod
    def clear_all():
        ImportTracker.imported_commands = {}

    @staticmethod
    def local_file_alias():
        return SmeltPath("local")

    @staticmethod
    def get_all_imported() -> Dict[SmeltPath, List["Command"]]:
        return ImportTracker.imported_commands


def try_get_target(target: "TargetRef") -> Optional["Target"]:
    tt = TempTarget.parse_string_smelt_target(
        target, ImportTracker.local_file_alias().path
    )
    try:
        return ImportTracker.imported_targets[tt.file_path][tt.name]
    except Exception as _:
        return None



from dataclasses import dataclass
import betterproto
from pysmelt.proto.smelt_telemetry import Event


@dataclass
class IsDoneSubscriber:
    """
    Simple subscriber that tells us if we're done executing
    """

    is_done: bool = False

    def process_message(self, message: Event):
        (variant, event_payload) = betterproto.which_one_of(message, "et")
        if variant == "invoke":
            (invoke_name, invoke_payload) = betterproto.which_one_of(
                event_payload, "InvokeVariant"
            )

            if invoke_name == "done":
                self.is_done = True
            else:
                pass
        else:
            pass

    def reset(self):
        self.is_done = False



from pysmelt.interfaces.procedural import import_as_target
from pysmelt.default_targets import test_group, raw_bash

for i in range(5):
    raw_bash(name=f"my_test_{i}", cmds=[f'echo "howdy partner from test {i}"'])
